@@ -66,10 +66,11 @@
 #define BASE_SLOT_ID   18UL
 
 static struct _Shared {
+	int initialized;
 	p11_dict *sessions;
 	p11_array *tokens;
 	char *paths;
-} gl = { NULL, NULL };
+} gl = { 0, NULL, NULL, NULL };
 
 /* Used during FindObjects */
 typedef struct _FindObjects {
@@ -133,6 +134,19 @@ lookup_object_inlock (p11_session *session,
 	return NULL;
 }
 
+static CK_RV
+check_index_writable (p11_session *session,
+                      p11_index *index)
+{
+	if (index == p11_token_index (session->token)) {
+		if (!p11_token_is_writable (session->token))
+			return CKR_TOKEN_WRITE_PROTECTED;
+		else if (!session->read_write)
+			return CKR_SESSION_READ_ONLY;
+	}
+
+	return CKR_OK;
+}
 
 static CK_RV
 lookup_slot_inlock (CK_SLOT_ID id,
@@ -181,6 +195,7 @@ create_tokens_inlock (p11_array *tokens,
 		const char *prefix;
 		const char *label;
 	} labels[] = {
+		{ "~/", "User Trust" },
 		{ DATADIR, "Default Trust" },
 		{ SYSCONFDIR, "System Trust" },
 		{ NULL },
@@ -284,10 +299,13 @@ sys_C_Finalize (CK_VOID_PTR reserved)
 	} else {
 		p11_lock ();
 
-			if (!gl.sessions) {
+			if (gl.initialized == 0) {
+				p11_debug ("trust module is not initialized");
 				rv = CKR_CRYPTOKI_NOT_INITIALIZED;
 
-			} else {
+			} else if (gl.initialized == 1) {
+				p11_debug ("doing finalization");
+
 				free (gl.paths);
 				gl.paths = NULL;
 
@@ -298,6 +316,11 @@ sys_C_Finalize (CK_VOID_PTR reserved)
 				gl.tokens = NULL;
 
 				rv = CKR_OK;
+				gl.initialized = 0;
+
+			} else {
+				gl.initialized--;
+				p11_debug ("trust module still initialized %d times", gl.initialized);
 			}
 
 		p11_unlock ();
@@ -310,7 +333,9 @@ sys_C_Finalize (CK_VOID_PTR reserved)
 static CK_RV
 sys_C_Initialize (CK_VOID_PTR init_args)
 {
-	CK_C_INITIALIZE_ARGS *args = NULL;
+	static const CK_C_INITIALIZE_ARGS def_args =
+		{ NULL, NULL, NULL, NULL, CKF_OS_LOCKING_OK, NULL, };
+	const CK_C_INITIALIZE_ARGS *args = NULL;
 	int supplied_ok;
 	CK_RV rv;
 
@@ -324,8 +349,9 @@ sys_C_Initialize (CK_VOID_PTR init_args)
 
 		rv = CKR_OK;
 
-		/* pReserved must be NULL */
 		args = init_args;
+		if (args == NULL)
+			args = &def_args;
 
 		/* ALL supplied function pointers need to have the value either NULL or non-NULL. */
 		supplied_ok = (args->CreateMutex == NULL && args->DestroyMutex == NULL &&
@@ -346,11 +372,17 @@ sys_C_Initialize (CK_VOID_PTR init_args)
 			rv = CKR_CANT_LOCK;
 		}
 
+		if (rv == CKR_OK && gl.initialized != 0) {
+			p11_debug ("trust module already initialized %d times",
+			           gl.initialized);
+
 		/*
 		 * We support setting the socket path and other arguments from from the
 		 * pReserved pointer, similar to how NSS PKCS#11 components are initialized.
 		 */
-		if (rv == CKR_OK) {
+		} else if (rv == CKR_OK) {
+			p11_debug ("doing initialization");
+
 			if (args->pReserved)
 				p11_argv_parse ((const char*)args->pReserved, parse_argument, NULL);
 
@@ -368,6 +400,8 @@ sys_C_Initialize (CK_VOID_PTR init_args)
 			}
 		}
 
+		gl.initialized++;
+
 	p11_unlock ();
 
 	if (rv != CKR_OK)
@@ -524,7 +558,7 @@ sys_C_GetTokenInfo (CK_SLOT_ID id,
 		info->firmwareVersion.minor = 0;
 		info->hardwareVersion.major = PACKAGE_MAJOR;
 		info->hardwareVersion.minor = PACKAGE_MINOR;
-		info->flags = CKF_TOKEN_INITIALIZED | CKF_WRITE_PROTECTED;
+		info->flags = CKF_TOKEN_INITIALIZED;
 		strncpy ((char*)info->manufacturerID, MANUFACTURER_ID, 32);
 		strncpy ((char*)info->model, TOKEN_MODEL, 16);
 		strncpy ((char*)info->serialNumber, TOKEN_SERIAL_NUMBER, 16);
@@ -546,6 +580,9 @@ sys_C_GetTokenInfo (CK_SLOT_ID id,
 			length = sizeof (info->label);
 		memset (info->label, ' ', sizeof (info->label));
 		memcpy (info->label, label, length);
+
+		if (!p11_token_is_writable (token))
+			info->flags |= CKF_WRITE_PROTECTED;
 	}
 
 	p11_unlock ();
@@ -587,8 +624,8 @@ sys_C_InitToken (CK_SLOT_ID id,
                  CK_ULONG pin_len,
                  CK_UTF8CHAR_PTR label)
 {
-	return_val_if_fail (check_slot (id), CKR_SLOT_ID_INVALID);
-	return_val_if_reached (CKR_TOKEN_WRITE_PROTECTED);
+	p11_debug ("not supported");
+	return CKR_FUNCTION_NOT_SUPPORTED;
 }
 
 static CK_RV
@@ -625,13 +662,16 @@ sys_C_OpenSession (CK_SLOT_ID id,
 		} else if (!(flags & CKF_SERIAL_SESSION)) {
 			rv = CKR_SESSION_PARALLEL_NOT_SUPPORTED;
 
-		} else if (flags & CKF_RW_SESSION) {
+		} else if ((flags & CKF_RW_SESSION) &&
+		           !p11_token_is_writable (token)) {
 			rv = CKR_TOKEN_WRITE_PROTECTED;
 
 		} else {
 			session = p11_session_new (token);
 			if (p11_dict_set (gl.sessions, &session->handle, session)) {
 				rv = CKR_OK;
+				if (flags & CKF_RW_SESSION)
+					session->read_write = true;
 				*handle = session->handle;
 				p11_debug ("session: %lu", *handle);
 			} else {
@@ -748,7 +788,8 @@ sys_C_InitPIN (CK_SESSION_HANDLE handle,
                CK_UTF8CHAR_PTR pin,
                CK_ULONG pin_len)
 {
-	return_val_if_reached (CKR_TOKEN_WRITE_PROTECTED);
+	p11_debug ("not supported");
+	return CKR_FUNCTION_NOT_SUPPORTED;
 }
 
 static CK_RV
@@ -758,7 +799,8 @@ sys_C_SetPIN (CK_SESSION_HANDLE handle,
               CK_UTF8CHAR_PTR new_pin,
               CK_ULONG new_pin_len)
 {
-	return_val_if_reached (CKR_TOKEN_WRITE_PROTECTED);
+	p11_debug ("not supported");
+	return CKR_FUNCTION_NOT_SUPPORTED;
 }
 
 static CK_RV
@@ -831,7 +873,8 @@ sys_C_CreateObject (CK_SESSION_HANDLE handle,
                     CK_OBJECT_HANDLE_PTR new_object)
 {
 	p11_session *session;
-	CK_BBOOL token;
+	p11_index *index;
+	CK_BBOOL val;
 	CK_RV rv;
 
 	return_val_if_fail (new_object != NULL, CKR_ARGUMENTS_BAD);
@@ -842,12 +885,15 @@ sys_C_CreateObject (CK_SESSION_HANDLE handle,
 
 		rv = lookup_session (handle, &session);
 		if (rv == CKR_OK) {
-			if (p11_attrs_findn_bool (template, count, CKA_TOKEN, &token) && token)
-				rv = CKR_TOKEN_WRITE_PROTECTED;
+			if (p11_attrs_findn_bool (template, count, CKA_TOKEN, &val) && val)
+				index = p11_token_index (session->token);
+			else
+				index = session->index;
+			rv = check_index_writable (session, index);
 		}
 
 		if (rv == CKR_OK)
-			rv = p11_index_add (session->index, template, count, new_object);
+			rv = p11_index_add (index, template, count, new_object);
 
 	p11_unlock ();
 
@@ -868,6 +914,7 @@ sys_C_CopyObject (CK_SESSION_HANDLE handle,
 	p11_session *session;
 	CK_ATTRIBUTE *original;
 	CK_ATTRIBUTE *attrs;
+	p11_index *index;
 	CK_BBOOL val;
 	CK_RV rv;
 
@@ -879,21 +926,22 @@ sys_C_CopyObject (CK_SESSION_HANDLE handle,
 
 		rv = lookup_session (handle, &session);
 		if (rv == CKR_OK) {
-			original = lookup_object_inlock (session, object, NULL);
+			original = lookup_object_inlock (session, object, &index);
 			if (original == NULL)
 				rv = CKR_OBJECT_HANDLE_INVALID;
 		}
 
 		if (rv == CKR_OK) {
-			if (p11_attrs_findn_bool (template, count, CKA_TOKEN, &val) && val)
-				rv = CKR_TOKEN_WRITE_PROTECTED;
+			if (p11_attrs_findn_bool (template, count, CKA_TOKEN, &val))
+				index = val ? p11_token_index (session->token) : session->index;
+			rv = check_index_writable (session, index);
 		}
 
 		if (rv == CKR_OK) {
 			attrs = p11_attrs_dup (original);
 			attrs = p11_attrs_buildn (attrs, template, count);
 			attrs = p11_attrs_build (attrs, &token, NULL);
-			rv = p11_index_take (session->index, attrs, new_object);
+			rv = p11_index_take (index, attrs, new_object);
 		}
 
 	p11_unlock ();
@@ -908,6 +956,9 @@ sys_C_DestroyObject (CK_SESSION_HANDLE handle,
                      CK_OBJECT_HANDLE object)
 {
 	p11_session *session;
+	CK_ATTRIBUTE *attrs;
+	p11_index *index;
+	CK_BBOOL val;
 	CK_RV rv;
 
 	p11_debug ("in");
@@ -916,11 +967,19 @@ sys_C_DestroyObject (CK_SESSION_HANDLE handle,
 
 		rv = lookup_session (handle, &session);
 		if (rv == CKR_OK) {
-			rv = p11_index_remove (session->index, object);
-			if (rv == CKR_OBJECT_HANDLE_INVALID) {
-				if (p11_index_lookup (p11_token_index (session->token), object))
-					rv = CKR_TOKEN_WRITE_PROTECTED;
+			attrs = lookup_object_inlock (session, object, &index);
+			if (attrs == NULL)
+				rv = CKR_OBJECT_HANDLE_INVALID;
+			else
+				rv = check_index_writable (session, index);
+
+			if (rv == CKR_OK && p11_attrs_find_bool (attrs, CKA_MODIFIABLE, &val) && !val) {
+				/* TODO: This should be replaced with CKR_ACTION_PROHIBITED */
+				rv = CKR_FUNCTION_REJECTED;
 			}
+
+			if (rv == CKR_OK)
+				rv = p11_index_remove (index, object);
 		}
 
 	p11_unlock ();
@@ -1030,6 +1089,9 @@ sys_C_SetAttributeValue (CK_SESSION_HANDLE handle,
                          CK_ULONG count)
 {
 	p11_session *session;
+	CK_ATTRIBUTE *attrs;
+	p11_index *index;
+	CK_BBOOL val;
 	CK_RV rv;
 
 	p11_debug ("in");
@@ -1038,11 +1100,30 @@ sys_C_SetAttributeValue (CK_SESSION_HANDLE handle,
 
 		rv = lookup_session (handle, &session);
 		if (rv == CKR_OK) {
-			rv = p11_index_set (session->index, object, template, count);
-			if (rv == CKR_OBJECT_HANDLE_INVALID) {
-				if (p11_index_lookup (p11_token_index (session->token), object))
-					rv = CKR_TOKEN_WRITE_PROTECTED;
+			attrs = lookup_object_inlock (session, object, &index);
+			if (attrs == NULL) {
+				rv = CKR_OBJECT_HANDLE_INVALID;
+			} else if (p11_attrs_find_bool (attrs, CKA_MODIFIABLE, &val) && !val) {
+				/* TODO: This should be replaced with CKR_ACTION_PROHIBITED */
+				rv = CKR_ATTRIBUTE_READ_ONLY;
 			}
+
+			if (rv == CKR_OK)
+				rv = check_index_writable (session, index);
+
+			/* Reload the item if applicable */
+			if (rv == CKR_OK && index == p11_token_index (session->token)) {
+				if (p11_token_reload (session->token, attrs)) {
+					attrs = p11_index_lookup (index, object);
+					if (p11_attrs_find_bool (attrs, CKA_MODIFIABLE, &val) && !val) {
+						/* TODO: This should be replaced with CKR_ACTION_PROHIBITED */
+						rv = CKR_ATTRIBUTE_READ_ONLY;
+					}
+				}
+			}
+
+			if (rv == CKR_OK)
+				rv = p11_index_set (index, object, template, count);
 		}
 
 	p11_unlock ();
